__author__  = "Andrea Gavana <andrea.gavana@gmail.com>, <gavana@kpo.kz>"
__date__    = "01 Apr 2007, 13:15 GMT"
__version__ = "0.1"
__docformat__ = "epytext"

# Start the imports
import os
import wx
import time

import sys
reload(sys)
sys.setdefaultencoding('utf-8')
del sys.setdefaultencoding

import wx.aui
import wx.lib.dialogs

# This is somehow needed by distutils, otherwise it bombs on Windows
# when you have py2App installed (!)
if wx.Platform == "__WXMAC__":
    try:
        import setuptools
    except:
        pass

# I need webbrowser for the help and tips and tricks
import webbrowser

# Let's import few modules I have written for GUI2Exe
from ProjectTreeCtrl import ProjectTreeCtrl
from MessageWindow import MessageWindow
from AUINotebookPage import AUINotebookPage
from DataBase import DataBase
from Project import Project
from Process import Process
from Widgets import CustomCodeViewer, Py2ExeMissing
from Utilities import opj, odict
from Constants import _auiImageList, _pywild, _defaultCompilers

# And import the fancy AdvancedSplash
import AdvancedSplash as AS

# I need that to have restorable perspectives
ID_FirstPerspective = wx.ID_HIGHEST + 10001

# It looks like that, while py2exe and py2app are installed on site-packages
# (or at least you can actually __import__ them), for the other 2 beasts
# is far more complicated, at least on Windows

def ImportCompilers():
    """ Simple function that imports the available compilers (if any). """
    
    compilers = odict()
    for compiler in _defaultCompilers:
        try:
            module = __import__(compiler)
            compilers[compiler] = module.__version__
            
        except ImportError:
            # No compiler, no party
            pass

    return compilers

# Ok, now get the compilers...
_compilers = ImportCompilers()



class GUI2Exe(wx.Frame):
    
    def __init__(self, parent, id=-1, title="", pos=wx.DefaultPosition, size=wx.DefaultSize,
                 style=wx.DEFAULT_FRAME_STYLE):
        """ Default wx.Frame class constructor. """
        
        wx.Frame.__init__(self, parent, id, title, pos, size, style)

        # Yes, I know, I am obsessively addicted to wxAUI
        self._mgr = wx.aui.AuiManager()
        self._mgr.SetManagedWindow(self)

        # Some default starting values for our class
        self.installDir = os.getcwd()      # where are we
        self.autoSave = False              # use autoSave? (not implemented yet...)
        self.deleteBuild = True            # delete the "build" directory (recommended)
        self.process = None                # keeps track of the compilation subprocess
        self.exeTimer = wx.Timer(self)     # I use that to monitor exe failures
        self.timerCount = 0                # same as above
        self.processTimer = wx.Timer(self) # used to monitor the external process
        
        self.perspectives = []

        # Create the status bar
        self.CreateBar()
        # Set frame properties (title, icon...)
        self.SetProperties()
        # Create the menu bar (lots of code, mostly always the same)
        self.CreateMenuBar()
        # Build the wx.aui.AuiNotebook image list
        # But why in the world is so different from wx.Notebook???
        self.BuildNBImageList()

        # Look if there already exists a database for GUI2Exe
        dbName = self.CheckForDatabase()

        # This is the left CustomTreeCtrl that holds all our projects
        self.projectTree = ProjectTreeCtrl(self)
        # This is the main window, the central pane
        self.mainPanel = wx.aui.AuiNotebook(self, -1, style=wx.aui.AUI_NB_DEFAULT_STYLE|
                                            wx.aui.AUI_NB_WINDOWLIST_BUTTON)
        # Let's be fancy and add a special logging window
        self.messageWindow = MessageWindow(self)
        # Call the database. This actually populates the project tree control
        self.dataBase = DataBase(self, dbName)

        # Add the panes to the wxAUI manager
        # Very nice the bug introduced in wxPython 2.8.3 about wxAUI Maximize buttons...
        self._mgr.AddPane(self.projectTree, wx.aui.AuiPaneInfo().Left().
                          Caption("GUI2Exe Projects").MinSize(wx.Size(200, -1)).
                          FloatingSize(wx.Size(200, 300)).Layer(1).MaximizeButton())
        self._mgr.AddPane(self.mainPanel, wx.aui.AuiPaneInfo().CenterPane())
        self._mgr.AddPane(self.messageWindow, wx.aui.AuiPaneInfo().Bottom().
                          Caption("Messages And Actions").MinSize(wx.Size(200, 100)).
                          FloatingSize(wx.Size(500, 300)).BestSize(wx.Size(200, size[1]/6)).
                          MaximizeButton())

        # Set all the flags for wxAUI         
        self.SetAllFlags()
        # Bind the main frame events
        self.BindEvents()

        # Save the current perspective to be reloaded later if requested
        self.perspectives.append(self._mgr.SavePerspective())
        # Update the wxAUI manager
        self._mgr.Update()
        # Read the default configuration file. At the moment it contains data
        # only for py2exe, and it might be moved to wx.Config
        self.ReadConfigurationFile()
        

    # ================================== #
    # GUI2Exe methods called in __init__ #
    # ================================== #
    
    def CreateBar(self):
        """ Creates the GUI2Exe status bar. """

        # Come on, let's see how fast the menubar is able to delete the text
        # I have in the status bar...
        self.statusBar = self.CreateStatusBar(2, wx.ST_SIZEGRIP)
        self.statusBar.SetStatusWidths([-1, -2])
        self.FillStatusBar()


    def MenuData(self):
        """ Handles all the information used to build the menu bar. """

        # That's really a bunch of data...
        
        return (("&File",
                    ("&New project...\tCtrl+N", "Add a new project to the project tree", "project", self.OnNewProject, ""),
                    ("Switch project &database...\tCtrl+D", "Load another GUI2Exe database file", "switch_db", self.OnSwitchDB, ""),
                    ("", "", "", "", ""),
                    ("&Save project\tCtrl+S", "Save the current project to database", "save_project", self.OnSaveProject, ""),
                    ("&Export setup file...\tCtrl+E", "Export the Setup.py file", "export_setup", self.OnExportSetup, ""),
                    ("", "", "", "", ""),
                    ("&Quit\tCtrl+Q", "Exit GUI2Exe", "exit", self.OnClose, "")),
                ("&Options",
                    ("Use &AutoSave", "AutoSaves your work every minute", "", self.OnAutoSave, wx.ITEM_CHECK),
                    ('De&lete "build" directory', "Delete the build folder at every compilation", "", self.OnDeleteBuild, wx.ITEM_CHECK),
                    ("", "", "", "", ""),
                    ("&Test executable\tCtrl+R", "Test the compiled file (if it exists)", "runexe", self.OnTestExecutable, ""),
                    ("Add &custom code...\tCtrl+U", "Add custom code to the setup script", "custom_code", self.OnCustomCode, "")), 
                ("&View",
                    ("&Setup script\tCtrl+P", "View the auto-generated setup script", "view_setup", self.OnViewSetup, ""),
                    ("", "", "", "", ""),
                    ("&Missing modules\tCtrl+M", "What the compiler thinks are the missing modules", "missingmodules", self.OnViewMissing, ""),
                    ("&Binary dependencies\tCtrl+B", "What the compiler says are the binary dependencies", "binarydependencies", self.OnViewMissing, ""),
                    ("", "", "", "", ""),
                    ("Compiler s&witches\tCtrl+W", "Show compilers switches and common options", "compiler_switches", self.OnCompilerSwitches, ""),
                    ("&Tips and tricks\tCtrl+T", "Show compilation tips and tricks", "tips_and_tricks", self.OnTipsAndTricks, "")),
                ("&Configuration",
                    ("Save &panes configuration...", "Save the current GUI panes configuration", "save_aui_config", self.OnSaveConfig, ""),
                    ("Restore original &GUI\tCtrl+G", "Restore the original GUI appearance", "restore_aui", self.OnRestorePerspective, "")),
                ("&Help",
                    ("GUI2Exe &help\tF1", "Opens the GUI2Exe help", "help", self.OnHelp, ""),
                    ("GUI2Exe &API\tF2", "Opens the GUI2Exe API reference", "api_reference", self.OnAPI, ""),
                    ("", "", "", "", ""),
                    ("Check for &upgrade\tF5", "Check for a GUI2Exe upgrade", "upgrade", self.OnCheckUpgrade, ""),
                    ("", "", "", "", ""),
                    ("&About GUI2Exe...", "About GUI2Exe and the Creator...", "about", self.OnAbout, "")))
                    
                    
    def CreateMenu(self, menuData):
        """ Creates a menu based on input menu data. """
        
        menu = wx.Menu()

        # Here is a bit trickier than what presented in Robin and Noel book,
        # but not that much.
        for eachLabel, eachStatus, eachIcon, eachHandler, eachKind in menuData:

            if not eachLabel:
                menu.AppendSeparator()
                continue

            # I need to find which menu holds the wxAUI-based "restore perspective"
            # as I have to bind on wx.EVT_MENU_RANGE with a specific start id 
            id = (eachLabel.find("Restore") >= 0 and [ID_FirstPerspective] or [-1])[0]
            # There are also few check menu items around...
            kind = (eachKind and [eachKind] or [wx.ITEM_NORMAL])[0]

            menuItem = wx.MenuItem(menu, id, eachLabel, eachStatus, kind=kind)
            if eachIcon:
                # Check menu items usually don't have associated icons
                menuItem.SetBitmap(self.CreateBitmap(eachIcon))

            menu.AppendItem(menuItem)
            if eachLabel.find('"build"') >= 0:
                # By default the "remove build directory" is on
                menuItem.Check(True)

            # Bind the event
            self.Bind(wx.EVT_MENU, eachHandler, menuItem)

        return menu


    def CreateMenuBar(self):
        """ Creates the main frame menu bar. """

        menuBar = wx.MenuBar()

        # loop over the bunch of data above
        for eachMenuData in self.MenuData():
            menuLabel = eachMenuData[0]
            menuItems = eachMenuData[1:]
            menuBar.Append(self.CreateMenu(menuItems), menuLabel)
    
        # Bind the special "restore perspective" menu item
        self.Bind(wx.EVT_MENU_RANGE, self.OnRestorePerspective, id=ID_FirstPerspective,
                  id2=ID_FirstPerspective+1000)

        # I need to keep track of this menu
        id = menuBar.FindMenu("Configuration")
        self.configMenu = menuBar.GetMenu(id)

        # We're done
        self.SetMenuBar(menuBar)
        

    def BuildNBImageList(self):
        """ Builds a fake image list for wx.aui.AuiNotebook. """

        # One day someone will explain why it doesn't handle wx.ImageList
        # like every other Book.
        self.nbImageList = []
        for png in _auiImageList:
            bmp = self.CreateBitmap(png)
            self.nbImageList.append(bmp)
            

    def SetProperties(self):
        """ Sets the main frame properties (title, icon...). """

        self.SetIcon(wx.Icon(opj(self.installDir + "/images/GUI2Exe.png"), wx.BITMAP_TYPE_PNG))
        self.SetTitle("GUI2Exe v" + __version__)


    def SetAllFlags(self):
        """ Sets all the fancy flags for wxAUI and friends. """

        # Allow to have active panes and transparent dragging
        self._mgr.SetFlags(self._mgr.GetFlags() ^ wx.aui.AUI_MGR_ALLOW_ACTIVE_PANE)
        self._mgr.SetFlags(self._mgr.GetFlags() ^ wx.aui.AUI_MGR_TRANSPARENT_DRAG)

        # Try to give a decent look to the gradients
        self._mgr.GetArtProvider().SetColor(wx.aui.AUI_DOCKART_INACTIVE_CAPTION_GRADIENT_COLOUR,
                                            wx.Colour(128, 128, 128))
        self._mgr.GetArtProvider().SetColor(wx.aui.AUI_DOCKART_ACTIVE_CAPTION_GRADIENT_COLOUR,
                                            wx.WHITE)

        # Very useful method
        self.mainPanel.SetUniformBitmapSize((16, 16))


    def CheckForDatabase(self):
        """ Checks if a database exists. If it doesn't, creates one anew. """

        # We build the database inside the user config folder, where we
        # also create a sub-directory called /.GUI2Exe
        standardPath = wx.StandardPaths.Get()
        configDir = opj(standardPath.GetUserConfigDir() + "/.GUI2Exe")
        configDb = opj(configDir + "/GUI2Exe_Database.db")
        
        if not os.path.isfile(configDb):
            # No database
            if not os.path.isdir(configDir):
                # And no directory. Create a new one.
                os.mkdir(configDir)

        return configDb.encode()
    

    def BindEvents(self):
        """ Binds all the events related to GUI2Exe. """

        self.Bind(wx.EVT_CLOSE, self.OnClose)
        # This one won't work in any case, I think
        self.Bind(wx.EVT_QUERY_END_SESSION, self.OnClose)
        # We monitor the external compilation process, when started
        self.Bind(wx.EVT_END_PROCESS, self.OnProcessEnded)
        self.Bind(wx.EVT_TIMER, self.OnExeTimer, self.exeTimer)
        # Bind the timer event. This allows us to monitor either the dry-run or
        # the real compilation using an external process which sends back to
        # us its output and error streams, and we monitor them in this event.
        self.Bind(wx.EVT_TIMER, self.OnProcessTimer, self.processTimer)

        # Let's do some fancy checking and painting during the page changing
        # an page closing event for wx.aui.AuiNotebook
        self.mainPanel.Bind(wx.aui.EVT_AUINOTEBOOK_PAGE_CLOSE, self.OnPageClosing)
        self.mainPanel.Bind(wx.aui.EVT_AUINOTEBOOK_PAGE_CHANGING, self.OnPageChanging)


    def ReadConfigurationFile(self):
        """ Reads the default configuration file (default project initialization). """

        # This should really be moved in wx.Config
        self.defaultConfig = {"py2exe": {}}
        fid = open(opj(self.installDir + "/defaultConfig.txt"), "rt")

        section = 0

        while 1:

            tline = fid.readline().encode()
            
            if not tline:
                fid.close()
                break

            if tline.find("%%") >= 0:
                # Ok, we found a compiler
                current = self.defaultConfig[tline.replace("%%", "").strip()]
                continue

            # I need the "--" separator to differentiate between string input
            # and other things
            if tline.find("--") >= 0:
                section += 1
                continue

            config, projectSwitches = tline.strip().split(":")

            if section:
                # it's not a string. I know eval is evil.
                projectSwitches = eval(projectSwitches)
            else:
                # it's a string, strip it
                projectSwitches = projectSwitches.strip()
                
            current[config] = projectSwitches
                

    # ============================== #
    # Event handlers for GUI2Exe     #
    # ============================== #
    
    def OnNewProject(self, event):
        """ A new project is being created. """

        # Send it to the project tree control
        self.projectTree.NewProject()


    def OnSwitchDB(self, event):
        """ Switch to another project database. """

        # Not implemented yet...
        event.Skip()


    def OnSaveProject(self, event):
        """ Saves the current project. """
        
        project = self.GetCurrentProject()
        if not project:
            # No page opened, you can't fool me
            return

        # Use an accessor function, as I need it also somewhere else below        
        self.SaveProject(project)

        
    def OnExportSetup(self, event):
        """ Exports the Setup.py file. """

        page = self.GetCurrentPage()
        if not page:
            # No page opened, you can't fool me
            return
        
        outputs = page.PrepareForCompile()
        if not outputs:
            # Setup.py file creation went wrong.
            # Have you set all the required variables?
            return

        # Compilation returns a big string containing the setup script and
        # the directory where the main script lives.
        setupScript, buildDir = outputs

        # Launch the save dialog        
        dlg = wx.FileDialog(self, message="Save file as ...", defaultDir=buildDir,
                            defaultFile="Setup.py", wildcard=_pywild,
                            style=wx.FD_SAVE|wx.FD_OVERWRITE_PROMPT)

        # Show the dialog and retrieve the user response. If it is the OK response, 
        # process the data.
        if dlg.ShowModal() == wx.ID_OK:
            path = dlg.GetPath()
            # Normally, at this point you would save your data using the file and path
            # data that the user provided to you.
            fp = file(path, 'w') # Create file anew
            fp.write(setupScript)
            fp.close()
            self.SendMessage("Message", "File %s successfully saved"%path)

        # Destroy the dialog. Don't do this until you are done with it!
        # BAD things can happen otherwise!
        dlg.Destroy()

        
    def OnClose(self, event):
        """ Handles the wx.EVT_CLOSE event for the main frame. """

        if self.exeTimer.IsRunning():
            # Kill the monitoring timer
            self.exeTimer.Stop()

        if self.process is not None:
            # Try to kill the process. Doesn't work very well
            self.process.Kill()
            self.processTimer.Stop()

        # Loop over all the opened wx.aui.AuiNotebook pages to see if
        # there are unsaved projects
        for pageNumber in xrange(self.mainPanel.GetPageCount()-1, -1, -1):
            if not self.HandlePageClosing(pageNumber, event):
                # User pressed cancel
                return
            
        # Un-initialize the wxAUI manager
        self._mgr.UnInit()
        # We're gone
        self.Destroy()


    def OnPageClosing(self, event):
        """ Handles the wx.aui.EVT_AUINOTEBOOK_PAGE_CLOSE event. """

        # Get the selected page
        selection = event.GetSelection()
        # Use the auxiliary method to handle this (is needed also in OnClose)
        self.HandlePageClosing(selection, event)
        

    def OnPageChanging(self, event):
        """ Handles the wx.aui.EVT_AUINOTEBOOK_PAGE_CHANGING event. """

        # Highlight the new item with a bold font
        newBook = self.mainPanel.GetPage(event.GetSelection())
        self.projectTree.HighlightItem(newBook.GetTreeItem(), True)
        
        if event.GetOldSelection() >= 0:
            # Restore the original font for the old item
            oldBook = self.mainPanel.GetPage(event.GetOldSelection())
            self.projectTree.HighlightItem(oldBook.GetTreeItem(), False)

        event.Skip()
        
        
    def OnAutoSave(self, event):
        """ Enables/Disables the AutoSave feature. """

        # Not implemented yet...
        self.autoSave = event.IsChecked()


    def OnDeleteBuild(self, event):
        """ Enables/Disables the automatic removal of the "build" folder. """

        # That's easy, we use it later
        self.deleteBuild = event.IsChecked()


    def OnTestExecutable(self, event):
        """ Test the compiled executable. """
        
        project = self.GetCurrentProject()
        if not project:
            # No page opened, you can't fool me
            return

        if not project.HasBeenCompiled():
            # The project hasn't been compiled yet
            msg = "This project has not been compiled yet."
            self.RunError("Error", msg)
            return

        # Try to run the successful compilation accessor method        
        self.SuccessfulCompilation(project, False)
        

    def OnCustomCode(self, event):
        """ Allows the user to add custom code to the Setup.py file. """

        project = self.GetCurrentProject()
        if not project:
            # No page opened, you can't fool me
            return

        currentPage = self.mainPanel.GetSelection()
        # Retrieve the existing custom code (if any)
        customCode = project.GetCustomCode()
        # Run the styled text control with the Python code in it, and
        # allow the user to edit it.
        frame = CustomCodeViewer(self, readOnly=False, text=customCode, project=project,
                                 page=currentPage)

        
    def OnViewSetup(self, event):
        """ Allows the user to see the Setup.py file. """

        wx.BeginBusyCursor()
        # I simply recreate the Setup.py script in memory
        self.RunCompile(view=True, run=False)
        wx.EndBusyCursor()


    def OnViewMissing(self, event):
        """ Shows the missing modules and dlls. """

        project = self.GetCurrentProject()
        if not project:
            # No page opened, you can't fool me
            return

        if not project.HasBeenCompiled():
            # The project hasn't been compiled yet
            msg = "This project has not been compiled yet."
            self.RunError("Error", msg)
            return

        wx.BeginBusyCursor()

        # Switch between the "show missing modules" and "show binary dependencies"        
        label = self.GetMenuBar().GetLabel(event.GetId())
        dll = (label.find("Binary") >= 0 and [True] or [False])[0]
        # Run the appropriate frame
        frame = Py2ExeMissing(self, project, dll)
        
        wx.EndBusyCursor()
        

    def OnCompilerSwitches(self, event):
        """ Shows the different compiler switches/options. """

        webbrowser.open_new(opj(self.installDir + "/docs/switches.html"))
        

    def OnTipsAndTricks(self, event):
        """ Shows the compilation tips and tricks. """

        webbrowser.open_new("http://www.py2exe.org/index.cgi/WorkingWithVariousPackagesAndModules")

        
    def OnSaveConfig(self, event):
        """ Saves the current GUI configuration, in terms of panes positions. """

        # Ask the user to enter a configuration name
        dlg = wx.TextEntryDialog(self, "Enter A Name For The New Configuration:",
                                 "Saving Panels Configuration")
        dlg.SetValue(("Perspective %d")%(len(self.perspectives)))
        
        if dlg.ShowModal() != wx.ID_OK:
            # No choice made, go back
            return

        value = dlg.GetValue()
        dlg.Destroy()

        if not value.strip():
            # Empty configuration name?
            self.RunError("Error", "Invalid perspective name!")
            return
        
        if len(self.perspectives) == 1:
            # Append a separator on the Configuration menu
            self.configMenu.AppendSeparator()

        # Append a new item in the Configuration menu
        item = wx.MenuItem(self.configMenu, ID_FirstPerspective + len(self.perspectives), value,
                           "Restore GUI configuration: %s"%value)
        item.SetBitmap(self.CreateBitmap("aui_config"))
        self.configMenu.AppendItem(item)

        # Save the current perspective        
        self.perspectives.append(self._mgr.SavePerspective())


    def OnRestorePerspective(self, event):
        """ Restore the selected GUI perspective. """

        self._mgr.LoadPerspective(self.perspectives[event.GetId() - ID_FirstPerspective])
        self._mgr.Update()


    def OnHelp(self, event):
        """ Shows the GUI2Exe help file. """

        # Not implemented yet...
        event.Skip()


    def OnAPI(self, event):
        """ Shows the GUI2Exe API help file. """

        webbrowser.open_new(opj(self.installDir + "/docs/api/index.html"))

        
    def OnCheckUpgrade(self, event):
        """ Checks for a possible upgrade of GUI2Exe. """

        # Not implemented yet...
        event.Skip()
        

    def OnAbout(self, event):
        """ Shows the about dialog for GUI2Exe. """

        msg = "This is the about dialog of GUI2Exe.\n\n" + \
              "Version %s"%__version__ + "\n"+ \
              "Author: Andrea Gavana @ 01 Apr 2007\n\n" + \
              "Please report any bug/request of improvements\n" + \
              "to me at the following addresses:\n\n" + \
              "andrea.gavana@gmail.com\ngavana@kpo.kz\n\n" + \
              "Thanks to Robin Dunn and the wxPython mailing list\n" + \
              "for the ideas and useful suggestions."
              
        self.RunError("Message", msg)


    def OnProcessTimer(self, event):
        """ Handles the wx.EVT_TIMER event for the main frame. """

        # This event runs only when a compilation process is alive
        # When the process finishes (or dies), we simply stop the timer
        if self.process is not None:
            # Defer to the Process class the message scanning
            self.process.HandleProcessMessages()
        

    def OnProcessEnded(self, event):
        """ Handles the wx.EVT_END_PROCESS for the main frame. """

        # We stop the timer that look for the compilation steps
        self.processTimer.Stop()

        # Handle the (eventual) remaning process messages
        self.process.HandleProcessMessages(True)
        # Destroy the process
        wx.CallAfter(self.process.Destroy)
        # Re-enable the buttons at the bottom
        self.messageWindow.EnableButtons(True)
 
        
    def OnExeTimer(self, event):
        """ Handles the wx.EVT_TIMER event for the main frame. """

        # Look if the log file exists
        self.timerCount += 1
        logFile = self.currentExe + ".log"
        if os.path.isfile(logFile):
            # log file is there. The executable crashed for some reason
            self.exeTimer.Stop()
            # Examine the log file
            self.ExamineLogFile(logFile)

        if self.timerCount > 200:
            # More than 20 seconds elapsed... the exe works or the machine
            # is fantastically slow. Reset everything
            self.exeTimer.Stop()
            self.currentExe = None
            self.timerCount = 0


    # ============================== #
    # Auxiliary methods for GUI2Exe  #
    # ============================== #

    def HandlePageClosing(self, selection, event):

        # Let's see if the project has been saved or not
        unSaved = self.mainPanel.GetPageBitmap(selection) == self.nbImageList[1]

        # Retrieve all the information we need before closing
        page = self.mainPanel.GetPage(selection)
        project = page.GetProject()
        treeItem = page.GetTreeItem()
        # Check if it is a close event or a wx.aui.AuiNotebookEvent
        isCloseEvent = isinstance(event, wx.CloseEvent)
            
        if not unSaved:
            # Mark the item as non-edited anymore (if it exists)
            self.projectTree.SetItemEditing(treeItem, False)
            if not isCloseEvent:
                event.Skip()
            return True

        # Not saved. If it wasn't ever saved before, not saving will delete
        # the item from the project tree
        msg = "Warning: the selected page contains unsaved data.\n\nDo you wish to save this project?"
        answer = self.RunError("Question", msg)

        if answer == wx.ID_CANCEL:
            # You want to think about it, eh?
            event.Veto()
            return False
        elif answer == wx.ID_YES:
            # Save the project, defer to the database
            self.dataBase.SaveProject(project)
            if not isCloseEvent:
                # We are handling wx.aui.AuiNotebook close event
                event.Skip()
        else:
            # Check if the project exists
            projectExists = self.dataBase.IsProjectExisting(project)
            if not isCloseEvent:
                event.Skip()

            if not projectExists:
                # Project is not in the database, so delete the tree item
                self.projectTree.DeleteProject([treeItem])
            else:
                # Mark the item as non-edited anymore (if it exists)
                self.projectTree.SetItemEditing(treeItem, False)       

        return True
            
        
    def SaveProject(self, project):
        """ Saves the current project. """

        # Send the data to the database        
        self.dataBase.SaveProject(project)
        # Visually indicates that the project has been saved
        self.UpdatePageBitmap(project.GetName(), 0, self.mainPanel.GetSelection())


    def UpdatePageBitmap(self, pageName, pageIcon, selection=None):
        """
        Updates the wx.aui.AuiNotebook page image and text, to reflect the
        current project state (saved/unsaved).
        """

        if selection is None:
            # Get the selection ourselves
            selection = self.mainPanel.GetSelection()

        if self.mainPanel.GetPageBitmap(selection) == self.nbImageList[pageIcon]:
            # No way, the page bitmap is the same
            return

        # Change the bitmap and the text        
        self.mainPanel.SetPageBitmap(selection, self.nbImageList[pageIcon])
        self.mainPanel.SetPageText(selection, pageName)
        

    def RunError(self, kind, msg, sendMessage=False):
        """
        An utility method that shows a message dialog with different functionalities
        depending on the input.
        """

        if sendMessage:
            # Send a message also to the log window at the bottom
            self.SendMessage(kind, msg.strip("."))
            
        if kind == "Message":    # is a simple message
            style = wx.OK | wx.ICON_INFORMATION
        elif kind == "Warning":  # is a warning
            style = wx.OK | wx.ICON_EXCLAMATION
        elif kind == "Question": # is a question
            style = wx.YES_NO | wx.CANCEL | wx.ICON_QUESTION
        else:                    # is an error
            style = wx.OK | wx.ICON_ERROR

        # Create the message dialog
        dlg = wx.MessageDialog(None, msg, "GUI2Exe %s"%kind, style|wx.STAY_ON_TOP)
        answer = dlg.ShowModal()
        dlg.Destroy()

        if kind == "Question":
            # return the answer, it was a question
            return answer


    def AddNewProject(self, treeItem, projectName):
        """ Adds a new project to the current center pane. """

        # Freeze all. It speeds up a bit the drawing
        busy = wx.BusyInfo("Creating new project...")    
        self.Freeze()

        # Create a new project, dictionary-based
        project = Project(self.defaultConfig, projectName)
        # The Project method adds a page to the center pane
        # I need it in another method as I use it also elsewhere below
        self.Project(project, treeItem, True)

        # Send a message to the log window at the bottom    
        self.SendMessage("Message", 'New project "%s" added'%projectName)

        # Time to warm up...        
        self.Thaw()
        del busy


    def LoadProject(self, treeItem, projectName):
        """ Loads a project in the center pane. """

        # Check if there already is a page with the same name
        page = self.IsAlreadyOpened(treeItem)
        if page is not None:
            # There is, bring the page into focus
            self.mainPanel.SetSelection(page)
            return
        
        # Load the project from the database
        project = self.dataBase.LoadProject(projectName)
        # The Project method adds a page to the center pane
        self.Project(project, treeItem, False)
        # Send a message to the log window at the bottom    
        self.SendMessage("Message", 'Project "%s" successfully loaded'%projectName)


    def RenameProject(self, treeItem, oldName, newName):
        """ The user has renamed the project in the project tree control. """

        # Rename the project in the database
        project = self.dataBase.RenameProject(oldName, newName)
        # Update information on screen (if needed)
        page = self.WalkAUIPages(treeItem)
        if page is None:
            # This page is not opened, go back
            return

        # Set the new name in the wx.aui.AuiNotebook tab
        self.mainPanel.SetPageText(page, newName)
        book = self.mainPanel.GetPage(page)
        if project is None:
            # Never saved in the database
            project = book.GetProject()
            project.SetName(newName)
            
        book.SetProject(project)
        # Update the label in the central panel
        book.GetPage(0).UpdateLabel(project.GetName(), project.GetCreationDate())
        
        
    def Project(self, project, treeItem, isNew):
        """ Auxiliary method used to actually add a page to the center pane. """

        # Get the project name        
        projectName = project.GetName()
        # Add a page to the wx.aui.AuiNotebook
        page = AUINotebookPage(self.mainPanel, project, _compilers)
        # Check project name and bitmap depending on the isNew state
        pageName = (isNew and [projectName+"*"] or [projectName])[0]
        bmp = (isNew and [1] or [0])[0]

        # Assing project and treeItem to the added page
        page.SetProject(project)
        page.SetTreeItem(treeItem)

        # Add the page to the wx.aui.AuiNotebook        
        self.mainPanel.AddPage(page, pageName, True, bitmap=self.nbImageList[bmp])
                

    def WalkAUIPages(self, treeItem):
        """ Walks over all the opened page in the center pane. """

        # Loop over all the wx.aui.AuiNotebook pages
        for indx in xrange(self.mainPanel.GetPageCount()):
            page = self.mainPanel.GetPage(indx)
            if page.GetTreeItem() == treeItem:
                # Yes, the page is already opened
                return indx

        # No page like that is there
        return None            
        

    def IsAlreadyOpened(self, treeItem):
        """ Looks if a page is already opened. """

        return self.WalkAUIPages(treeItem)
    
    
    def CloseAssociatedPage(self, treeItem):
        """
        A method used to close a wx.aui.AuiNotebook page when an item in the
        project tree is deleted.
        """

        page = self.WalkAUIPages(treeItem)
        if page is not None:
            self.mainPanel.DeletePage(page)


    def ReassignPageItem(self, oldItem, newItem):
        """ Reassigns an item to an opened page after a drag and drop operation. """

        page = self.WalkAUIPages(oldItem)
        if page is not None:
            # We found an opened page
            page = self.mainPanel.GetPage(page)
            page.SetTreeItem(newItem)


    def FillStatusBar(self):
        """ Fills the statusbar fields with different information. """

        # Get the wxPython version information
        wxPythonVersion = wx.version()
        statusText = "wxPython %s"%wxPythonVersion

        for compiler, version in _compilers.items():
            # adds the compilers information found
            statusText += ", %s %s"%(compiler, version)

        # Ah, by the way, thank you menu bar for deleting my status bar messages...            
        self.statusBar.SetStatusText(statusText, 1)
        self.statusBar.SetStatusText("Welcome to GUI2Exe", 0)


    def CreateBitmap(self, fileName):
        """ Utility function to create bitmap passing a filename. """
        
        bmp = wx.Bitmap(opj(self.installDir + "/images/%s.png"%fileName), wx.BITMAP_TYPE_PNG)
        return bmp
    

    def GetCurrentPage(self):
        """ Returns the current LabelBook page. """

        book = self.GetCurrentBook()
        if not book:
            # No page opened, you can't fool me
            return
        
        return book.GetPage(book.GetSelection())


    def GetCurrentBook(self):
        """ Returns the current wx.aui.AuiNotebook page (a LabelBook). """

        if self.mainPanel.GetPageCount() == 0:
            # No page opened, fire an error
            msg = "No project has been loaded"
            self.RunError("Error", msg +".", True)
            return None

        # Return the current page (is a LabelBook)
        selection = self.mainPanel.GetSelection()
        return self.mainPanel.GetPage(selection)
        

    def GetCurrentProject(self):
        """ Returns the current project associated to a LabelBook. """

        book = self.GetCurrentBook()
        if not book:
            # No page opened, you can't fool me
            return
        
        return book.GetProject()
    

    def RunCompile(self, view, run):
        """
        Auxiliary method. Depending on the input, it does different things:
        - view=True ==> Get the Setup.py script and displays it (run is discarded);
        - run=False ==> Start a dry-run (a compilation that does nothing);
        - run=True  ==> Start the real compilation process.
        """
    
        page = self.GetCurrentPage()
        if not page:
            # No page opened, you can't fool me
            return
        
        outputs = page.PrepareForCompile()
        if not outputs:
            # Setup.py file creation went wrong.
            # Have you set all the required variables?
            return

        setupScript, buildDir = outputs
    
        if view:    # we just want to view the code
            frame = CustomCodeViewer(self, readOnly=True, text=setupScript)
            return
        
        # Disable the run buttons
        self.messageWindow.EnableButtons(False)

        # Get all the information we need about this project        
        compiler = page.GetName()
        project = self.GetCurrentProject()
        currentPage = self.mainPanel.GetSelection()

        # Start the external process, which is actually subclassed in
        # The Process class
        self.process = Process(self, buildDir, setupScript, run, compiler,
                               project, currentPage)

        # Start the monitoring timer
        self.processTimer.Start(500)
        # Start the process
        self.process.Start()


    def KillCompile(self):
        """ Kills (or tries to) the compilation process. """

        if self.process:
            # Stop the process timer
            self.processTimer.Stop()
            # This doesn't work, at least on Windows.
            # The process is still there and there is no os.kill on Windows
            self.process.Kill()
            # Send a failure message to the log window at the bottom
            self.SendMessage("Warning", "Compilation killed by user")
            # Re-enable the run buttons
            self.messageWindow.EnableButtons(True)
            

    def SuccessfulCompilation(self, project, ask=True):
        """
        Assumes that the compilation process was successful and tries to test
        the new exe file.
        """

        if ask:
            # We came from an wx.EVT_END_PROCESS event
            msg = "The compiler has successfully built your executable.\n" \
                  "Do you wish to test your application?"
            answer = self.RunError("Question", msg)
            if answer != wx.ID_YES:
                return

        # Get the executable name from the Project class
        exeName = project.GetExecutableName("py2exe")
        if not os.path.isfile(exeName):
            # No such file, have you compiled it?
            msg = "This project has never been compiled or its executable has been deleted."
            self.RunError("Error", msg)
            return

        # Starts the compiled exe file
        msg = "Starting compiled executable..."
        busy = wx.BusyInfo(msg)
        self.SendMessage("Message", msg) 

        logFile = exeName + ".log"
        # Remove the log file or it will fool us later
        if os.path.isfile(logFile):
            try:
                os.remove(logFile)
            except:
                pass
            
        self.currentExe = exeName
        # Start the monitoring timer to check if a log file is created
        self.exeTimer.Start(100)
        os.spawnl(os.P_NOWAIT, exeName)

        del busy
        wx.SafeYield()


    def ExamineLogFile(self, logFile):
        """ Examine a log file, created by an executable that crashed fpr some reason. """

        # Reset few variables, we don't need them anymore
        self.currentExe = None
        self.timerCount = 0

        # Send the error message to the log window at the bottom
        self.SendMessage("Error", "Executable terminated with errors or warnings")
        
        # Ask the user if he/she wants to examine the tracebacks in the log file
        msg = "It appears that the executable generated errors in file:\n\n%s" % logFile
        msg += "\n\nDo you want to examine the tracebacks?"
        answer = self.RunError("Question", msg)
        if answer != wx.ID_YES:
            return

        wx.BeginBusyCursor()

        # Read the log file        
        fid = open(logFile, "rt")
        msg = fid.read()
        fid.close()
        # And displays it in a scrolled message dialog
        dlg = wx.lib.dialogs.ScrolledMessageDialog(self, msg, "Tracebacks in log file")
        wx.EndBusyCursor()
        dlg.ShowModal()
        dlg.Destroy()

        
    def SendMessage(self, kind, message):
        """ Sends a message to the log window at the bottom. """

        self.messageWindow.SendMessage(kind, message)
        

    def GetVersion(self):
        """ Return the current GUI2Exe version. """

        return __version__
    

class GUI2ExeSplashScreen(AS.AdvancedSplash):

    def __init__(self, app):
        """ A fancy splash screen :-D """

        bmp = wx.Bitmap("images/gui2exe_splash.png", wx.BITMAP_TYPE_PNG)
        AS.AdvancedSplash.__init__(self, None, bitmap=bmp, timeout=5000,
                                   extrastyle=AS.AS_TIMEOUT|AS.AS_CENTER_ON_SCREEN)
        
        self.Bind(wx.EVT_CLOSE, self.OnClose)
        self.fc = wx.FutureCall(2000, self.ShowMain)
        self.app = app


    def OnClose(self, evt):
        """ Handles the wx.EVT_CLOSE event for GUI2ExeSplashScreen. """
        
        # Make sure the default handler runs too so this window gets
        # destroyed
        evt.Skip()
        self.Hide()
        
        # if the timer is still running then go ahead and show the
        # main frame now
        if self.fc.IsRunning():
            # Stop the wx.FutureCall timer
            self.fc.Stop()
            self.ShowMain()


    def ShowMain(self):
        """ Shows the main application (GUI2Exe). """

        # Get the size of the display
        size = wx.GetDisplaySize()
        # We run at 5/6 of that size
        xvideo, yvideo = 5*size.x/6, 5*size.y/6
        frame = GUI2Exe(None, -1, "", size=(xvideo, yvideo))

        self.app.SetTopWindow(frame)
        frame.CenterOnScreen()
        frame.Show()

        if self.fc.IsRunning():
            self.Raise()
            
            
class GUI2ExeApp(wx.App):
    
    def OnInit(self):

        wx.SetDefaultPyEncoding("utf-8")
        
        splash = GUI2ExeSplashScreen(self)
        splash.Show()

        self.SetAppName("GUI2Exe")

        return True
        

if __name__ == "__main__":
    app = GUI2ExeApp(0)
    app.MainLoop()

